; The static (or initialisation buffer) i.e. stuff that doesn't change for each
; time this code is called.
Scales		.assign 0
LightCol	.assign 1
LightPos	.assign 2
CamPos		.assign 3
Transform	.assign 8 
LightTrans	.assign 12

; The input buffer (relative the the start of one of the double buffers)
NumVerts	.assign 0 
GifPacket  	.assign	1 
UVStart		.assign 2
NormStart	.assign 3
VertStart	.assign 4

; The output buffer (relative the the start of one of the double buffers)
GifPacketOut	.assign 253
UVStartOut		.assign 254
NormStartOut	.assign 255
VertStartOut	.assign 256


; Note that we have 4 data buffers: InputBuffer0, OutputBuffer0, InputBuffer1, and OutputBuffer1.
; Each buffer is 252 quad words. The different buffers are selected by reading the current offset
; from xtop (which is swapped automatically by the PS2 after each MSCALL / MSCNT).


.include "vcl_sml.i"

.init_vf_all
.init_vi_all
.syntax new

.vu

--enter
--endenter

; The START or Init code, that is only called once per frame.
START:
	fcset		0x000000

	lq			fTransform[0], Transform+0(vi00)
	lq			fTransform[1], Transform+1(vi00)
	lq			fTransform[2], Transform+2(vi00)
	lq			fTransform[3], Transform+3(vi00)
	lq			fLightTrans[0], LightTrans+0(vi00)
	lq			fLightTrans[1], LightTrans+1(vi00)
	lq			fLightTrans[2], LightTrans+2(vi00)
	lq			fLightTrans[3], LightTrans+3(vi00)
	lq			fScales, Scales(vi00)
	
	lq			LightPos, LightPos(vi00)
	MatrixMultiplyVertex LightPos, fLightTrans, LightPos	; Transform the light position into world space

; This begin code is called once per batch
begin:
	xtop		iDBOffset		; Load the address of the current buffer (will either be QW 16 or QW 520)
	ilw.x		iNumVerts, NumVerts(iDBOffset)
	iadd		iNumVerts, iNumVerts, iDBOffset
	iadd		Counter, vi00, iDBOffset

loop:
	lq			fVert, VertStart(Counter)		; Load the vertex from the input buffer
	mul         acc,  fTransform[0],  fVert[x]	; Transform it
    madd        acc,  fTransform[1],  fVert[y]
    madd        acc,  fTransform[2],  fVert[z]
    madd        Vert, fTransform[3],  fVert[w]
    clipw.xyz	Vert, Vert						; Clip it
	fcand		vi01, 0x3FFFF
	iaddiu		iADC, vi01, 0x7FFF
	ilw.w		iNoDraw, UVStart(Counter)		; Load the iNoDraw flag. If true we should set the ADC bit so the vert isn't drawn
	iadd		iADC, iADC, iNoDraw
	isw.w		iADC, VertStartOut(Counter)
	div         q,    vf00[w], Vert[w]			
	lq			UV,   UVStart(Counter)			; Handle the tex-coords
	mul			UV,   UV, q
	sq			UV,   UVStartOut(Counter)
	mul.xyz     Vert, Vert, q					; Scale the final vertex to fit to the screen.
	mula.xyz	acc, fScales, vf00[w]
	madd.xyz	Vert, Vert, fScales
	ftoi4.xyz	Vert, Vert
	sq.xyz		Vert, VertStartOut(Counter)		; And store in the output buffer
	
	; DIFFUSE LIGHTING
	lq			Norm, NormStart(Counter)				; Load the normal
	MatrixMultiplyVertex Norm, fLightTrans, Norm		; Transform the normal into world space (important that norm.w is 0)
	MatrixMultiplyVertex Vert, fLightTrans, fVert		; Transform the vertex into world space
	sub.xyz		ToLight, LightPos, Vert					; Get the vector from the vert to the light
	VectorNormalizeXYZ ToLight, ToLight					; Normalise this vector (and it becomes the light direction, just as in directional lighting)
	VectorDotProduct  fIntensity, ToLight, Norm			; Get the intensity
	
	; SPECULAR
	lq.xyz		Cam, CamPos(vi00)
	sub.xyz		Cam, Cam, Vert
	VectorNormalizeXYZ Cam, Cam
	add.xyz		HalfVec, Cam, ToLight
	VectorNormalizeXYZ HalfVec, HalfVec
	VectorDotProduct  fSpecIntensity, HalfVec, Norm
	;max.x		fSpecIntensity, fSpecIntensity, vf00			; Clamp to > 0
	mul.x		fSpecIntensity, fSpecIntensity, fSpecIntensity	; Square it
	mul.x		fSpecIntensity, fSpecIntensity, fSpecIntensity	; 4th power
	mul.x		fSpecIntensity, fSpecIntensity, fSpecIntensity	; 8th power
	mul.x		fSpecIntensity, fSpecIntensity, fSpecIntensity	; 16th power
	mul.x		fSpecIntensity, fSpecIntensity, fSpecIntensity	; 32nd power
	add.x		fIntensity, fIntensity, fSpecIntensity	; Get the total intensity
	
	max.x		fIntensity, fIntensity, vf00			; Clamp to > 0
	mini.x		fIntensity, fIntensity, vf00[w]			; Clamp to < 1
	
	lq			fLightCol, LightCol(vi00)				; Load the light colour
	mul.xyz		fLightCol, fLightCol, fIntensity[x]		; Scale the colour by the intensity
	loi			255										
	mul			fLightCol, fLightCol, i					; Scale the values from 0-1 to the range 0-127
    ftoi0   	iLightCol, fLightCol					; Convert to ints
	sq   		iLightCol, NormStartOut(Counter)		; And write to the output buffer
	
	iaddiu		Counter, Counter, 3
	ibne		Counter, iNumVerts, loop				; Loop until all of the verts in this batch are done.
	iaddiu		iKick, iDBOffset, GifPacketOut
	lq			GP, GifPacket(iDBOffset)				; Copy the GIFTag to the output buffer
	sq			GP, GifPacketOut(iDBOffset)
	xgkick		iKick									; and render!
	
--cont
														; --cont is like end, but it really means pause, as this is where the code
														; will pick up from when MSCNT is called.
	b			begin									; Which will make it hit this code which takes it back to the start, but
														; skips the initialisation which we don't want done twice.

--exit
--endexit
